package net.sergeych.bipack

import kotlinx.serialization.SerializationStrategy
import kotlinx.serialization.descriptors.SerialDescriptor
import kotlinx.serialization.encoding.AbstractEncoder
import kotlinx.serialization.encoding.CompositeEncoder
import kotlinx.serialization.modules.EmptySerializersModule
import kotlinx.serialization.modules.SerializersModule
import kotlinx.serialization.serializer
import net.sergeych.bintools.*

class BipackEncoder(val output: DataSink) : AbstractEncoder() {
    override val serializersModule: SerializersModule = EmptySerializersModule
    override fun encodeBoolean(value: Boolean) = output.writeByte(if (value) 1 else 0)
    override fun encodeByte(value: Byte) = output.writeByte(value.toInt())
    override fun encodeShort(value: Short) = output.writeNumber(value.toInt())
    override fun encodeInt(value: Int) = output.writeNumber(value)

    fun encodeUInt(value: UInt) = output.writeNumber(value)
    override fun encodeLong(value: Long) = output.writeNumber(value)
    override fun encodeFloat(value: Float) = output.writeNumber(value)
    override fun encodeDouble(value: Double) = output.writeI64(value.toRawBits())
    override fun encodeChar(value: Char) = output.writeNumber(value.code.toUInt())
    override fun encodeString(value: String) {
//        output.writeUTF(value)
        writeBytes(value.encodeToByteArray())
    }

    fun writeBytes(value: ByteArray) {
        output.writeNumber(value.size.toUInt())
        output.writeBytes(value)
    }

    override fun encodeEnum(enumDescriptor: SerialDescriptor, index: Int) = output.writeNumber(index.toUInt())

    override fun beginCollection(descriptor: SerialDescriptor, collectionSize: Int): CompositeEncoder {
        encodeUInt(collectionSize.toUInt())
        return this
    }

    override fun beginStructure(descriptor: SerialDescriptor): CompositeEncoder {
        for( a in descriptor.annotations) {
            if (a is Framed) {
                output.writeU32(
                    CRC.crc32(descriptor.serialName.encodeToByteArray())
                )
            }
            else if( a is ExtendableFormat) {
                encodeUInt(descriptor.elementsCount.toUInt())
            }
        }
        return super.beginStructure(descriptor)
    }

    override fun encodeNull() = encodeBoolean(false)
    override fun encodeNotNullMark() = encodeBoolean(true)

    companion object {
        fun <T> encode(serializer: SerializationStrategy<T>, value: T,sink: DataSink) {
            val encoder = BipackEncoder(sink)
            encoder.encodeSerializableValue(serializer, value)
        }

        fun <T> encode(serializer: SerializationStrategy<T>, value: T): ByteArray =
            ArrayDataSink().also { encode(serializer, value, it)}.toByteArray()

        inline fun <reified T> encode(value: T) = encode(serializer(), value)
        inline fun <reified T> encode(value: T,sink: DataSink) = encode(serializer(), value, sink)

    }
}
